import pytest
from cirq import Circuit as CirqCircuit
from cirq import CNOT, H, X, LineQubit, NamedQubit
from cirq.testing import assert_same_circuits
from pytket import Circuit as TketCircuit
from qiskit import QuantumCircuit as QiskitCircuit
from qiskit import transpile as qiskit_transpile
from qiskit.converters import circuit_to_dag
from qiskit.quantum_info import Statevector
from qiskit.transpiler.passes import GatesInBasis, CountOps
from qiskit.transpiler.passes.utils import CheckMap
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.circuit.library import HGate, XGate
from ucc.tests.mock_backends import Mybackend
from ucc import compile
from ucc.transpilers.ucc_defaults import UCCDefault1
from ucc.transpilers.aqc.mps_pass import MPSPass
import numpy as np


def random_area_law_circuit(N, seed=12345):
    """A circuit to generate a random area-law statevector.

    Parameters:
        N (int): Number of qubits

    Returns:
        QiskitCircuit: Output circuit
    """
    np.random.seed(seed)

    state = np.random.rand(2**N) + 1j * np.random.rand(2**N)
    state /= np.linalg.norm(state)

    circuit = QiskitCircuit(N)
    circuit.initialize(state, range(N))

    return circuit


def qcnn_circuit(N, seed=12345):
    """A circuit to generate a Quantum Convolutional Neural Network

    Parameters:
        N (int): Number of qubits

    Returns:
        QiskitCircuit: Output circuit
    """
    rng = np.random.default_rng(seed=seed)

    qc = QiskitCircuit(N)
    num_layers = int(np.ceil(np.log2(N)))
    i_conv = 0
    for i_layer in range(num_layers):
        for i_sub_layer in [0, 2**i_layer]:
            for i_q1 in range(i_sub_layer, N, 2 ** (i_layer + 1)):
                i_q2 = 2**i_layer + i_q1
                if i_q2 < N:
                    qc.rxx(rng.random(), i_q1, i_q2)
                    qc.ry(rng.random(), i_q1)
                    qc.ry(rng.random(), i_q2)
                    i_conv += 1

    return qc


def random_clifford_circuit(num_qubits, seed=12345):
    """Generate a random clifford circuit
    Parameters:
        num_qubits (int): Number of qubits
        seed (int): Optional. Seed the random number generator, default=12345

    Returns:
        QuantumCircuit: Clifford circuit
    """
    # This code is used to generate the QASM file
    from qiskit.circuit.random import random_clifford_circuit

    gates = ["cx", "cz", "cy", "swap", "x", "y", "z", "s", "sdg", "h"]
    qc = random_clifford_circuit(
        num_qubits,
        gates=gates,
        num_gates=10 * num_qubits * num_qubits,
        seed=seed,
    )
    return qc


def test_qiskit_compile():
    circuit = QiskitCircuit(2)
    circuit.h(0)
    circuit.cx(0, 1)
    result_circuit = compile(circuit, return_format="original")
    assert isinstance(result_circuit, QiskitCircuit)


def test_cirq_compile():
    qubits = LineQubit.range(2)
    circuit = CirqCircuit(H(qubits[0]), CNOT(qubits[0], qubits[1]))
    result_circuit = compile(circuit, return_format="original")
    assert isinstance(result_circuit, CirqCircuit)


def test_tket_compile():
    circuit = TketCircuit(2)
    circuit.H(0)
    circuit.CX(0, 1)
    result_circuit = compile(circuit, return_format="original")
    assert isinstance(result_circuit, TketCircuit)


def test_callback():
    was_called = False

    def my_callback(**kwargs):
        nonlocal was_called
        was_called = True

    circuit = QiskitCircuit(2)
    circuit.h(0)
    circuit.cx(0, 1)
    _ = compile(circuit, callback=my_callback)
    assert was_called


def test_custom_pass():
    """Verify that a custom pass works with a non-qiskit input circuit"""

    class HtoX(TransformationPass):
        """Toy transformation that converts all H gates to X gates"""

        def run(self, dag):
            for node in dag.op_nodes():
                if isinstance(node.op, HGate):
                    dag.substitute_node(node, XGate())
            return dag

    # Example usage with a cirq circuit, stil showcasing the cross-frontend compatibility

    qubit = NamedQubit("q_0")
    cirq_circuit = CirqCircuit(H(qubit))

    post_compiler_circuit = compile(cirq_circuit, custom_passes=[HtoX()])
    assert_same_circuits(post_compiler_circuit, CirqCircuit(X(qubit)))


def test_compile_target_backend_opset():
    circuit = QiskitCircuit(3)
    circuit.cz(0, 1)
    circuit.cz(0, 2)

    # Create a simple backend that does not have direct CX between 0 and 2
    t = Mybackend()
    # Check that the gates in the original circuit are not support by the target
    # to ensure this isn't a trival check
    assert set(op.name for op in circuit).issubset(t.operation_names) is False

    result_circuit = compile(
        circuit, return_format="original", target_backend=t
    )
    # Check that the gates in the final circuit are all supported on the target device
    assert set(op.name for op in result_circuit).issubset(t.operation_names)


def test_compile_target_backend_coupling_map():
    circuit = QiskitCircuit(3)
    circuit.cx(0, 1)
    circuit.cx(0, 2)

    # Create a simple target that does not have direct CX between 0 and 2
    t = Mybackend()
    result_circuit = compile(
        circuit, return_format="original", target_backend=t
    )
    # Check that the compiled circuit respects the coupling map of the target device
    analysis_pass = CheckMap(
        t.target.build_coupling_map(), property_set_field="check_map"
    )

    dag = circuit_to_dag(result_circuit)
    analysis_pass.run(dag)
    assert analysis_pass.property_set["check_map"]


def test_compile_with_no_target_gateset_or_device():
    """Test that the final circuit is in the default gateset if no `target_gateset` or `target_backend` is provided."""

    # Circuit not in the default target_gateset {"cx", "rz", "rx", "ry", "h"}
    circuit = QiskitCircuit(2)
    circuit.cz(0, 1)
    circuit.h(0)

    result_circuit = compile(
        circuit,
    )

    assert set(op.name for op in result_circuit).issubset(
        {"cx", "rz", "rx", "ry", "h"}
    )

    # Circuit already in the default target_gateset {"cx", "rz", "rx", "ry", "h"}
    circuit = QiskitCircuit(2)
    circuit.cx(0, 1)
    circuit.h(0)

    result_circuit = compile(
        circuit,
    )

    assert set(op.name for op in result_circuit).issubset(
        {"cx", "rz", "rx", "ry", "h"}
    )


def test_raise_error_on_bad_backend():
    circuit = QiskitCircuit(2)
    circuit.cx(0, 1)
    circuit.h(0)

    class BadBackend:
        pass

    with pytest.raises(ValueError):
        _ = compile(circuit, target_backend=BadBackend())


def test_bqskit_compile():
    from ucc.transpilers.ucc_bqskit import BQSKitTransformationPass

    bqskit_pass = BQSKitTransformationPass()
    qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[3];
        h q[0];
        cp(1.5707963267948966) q[1], q[0];
        h q[1];
        cp(0.7853981633974483) q[2], q[0];
        cp(1.5707963267948966) q[2], q[1];
        h q[2];
        swap q[0], q[2];
        h q[0];
        cp(-1.5707963267948966) q[1], q[0];
        h q[1];
        cp(-0.7853981633974483) q[2], q[0];
        cp(-1.5707963267948966) q[2], q[1];
        h q[2];
        swap q[0], q[2];
        """
    # This qasm describes a 3-qubit QFT followed by a 3-qubit inverse QFT.
    # This circuit resolves to the identity, but that's not obvious to
    # most synthesis tools.
    # BQSKit using LEAP will usually remove all 2-qubit gates
    # from the circuit, leaving 3 u3 gates that don't do anything
    # because it just focuses on the 2-qubit gates. A further
    # post processing step would remove these 1-qubit gates, but in
    # more realistic use cases, its often not worth the extra processing.

    def get_post_cx_count(circuit, custom_passes=[]):
        post_compiler_circuit = compile(qasm, custom_passes=custom_passes)
        analysis_pass = CountOps()
        dag = circuit_to_dag(
            QiskitCircuit.from_qasm_str(post_compiler_circuit)
        )
        analysis_pass.run(dag)
        if "cx" in analysis_pass.property_set["count_ops"]:
            return analysis_pass.property_set["count_ops"]["cx"]
        else:
            return 0

    assert get_post_cx_count(qasm, [bqskit_pass]) < get_post_cx_count(qasm)


@pytest.mark.parametrize("N", [5, 8, 10, 11])
def test_compile_with_mps_pass(N):
    """Test that the circuit compiled by `MPSPass` works as expected."""
    circuit = random_area_law_circuit(N)
    circuit = qiskit_transpile(circuit, basis_gates=["u3", "cx"])

    compiled_circuit = compile(
        circuit, target_gateset=["u3", "cx"], custom_passes=[MPSPass()]
    )

    # Check that the fidelity is above 0.9
    fidelity = np.vdot(
        Statevector(circuit).data, Statevector(compiled_circuit).data
    )

    assert np.abs(fidelity) > 0.9

    # Assert circuit depth and number of cx gates is lower
    assert circuit.depth() > compiled_circuit.depth()
    assert circuit.count_ops().get("cx", 0) > compiled_circuit.count_ops().get(
        "cx", 0
    )


def test_compile_trivial_state_with_mps_pass():
    """Test that `MPSPass` does not use CX gates when state has no entanglement."""
    qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[10];
        ry(pi/2) q[9];
        rx(pi) q[9];
        rz(pi/4) q[9];
        cx q[9],q[8];
        rz(-pi/4) q[8];
        cx q[9],q[8];
        rz(pi/4) q[8];
        ry(pi/2) q[8];
        rx(pi) q[8];
        rz(pi/4) q[8];
        rz(pi/8) q[9];
        cx q[9],q[7];
        rz(-pi/8) q[7];
        cx q[9],q[7];
        rz(pi/8) q[7];
        cx q[8],q[7];
        rz(-pi/4) q[7];
        cx q[8],q[7];
        rz(pi/4) q[7];
        ry(pi/2) q[7];
        rx(pi) q[7];
        rz(pi/4) q[7];
        rz(pi/8) q[8];
        rz(pi/16) q[9];
        cx q[9],q[6];
        rz(-pi/16) q[6];
        cx q[9],q[6];
        rz(pi/16) q[6];
        cx q[8],q[6];
        rz(-pi/8) q[6];
        cx q[8],q[6];
        rz(pi/8) q[6];
        cx q[7],q[6];
        rz(-pi/4) q[6];
        cx q[7],q[6];
        rz(pi/4) q[6];
        ry(pi/2) q[6];
        rx(pi) q[6];
        rz(pi/4) q[6];
        rz(pi/8) q[7];
        rz(pi/16) q[8];
        rz(pi/32) q[9];
        cx q[9],q[5];
        rz(-pi/32) q[5];
        cx q[9],q[5];
        rz(pi/32) q[5];
        cx q[8],q[5];
        rz(-pi/16) q[5];
        cx q[8],q[5];
        rz(pi/16) q[5];
        cx q[7],q[5];
        rz(-pi/8) q[5];
        cx q[7],q[5];
        rz(pi/8) q[5];
        cx q[6],q[5];
        rz(-pi/4) q[5];
        cx q[6],q[5];
        rz(pi/4) q[5];
        ry(pi/2) q[5];
        rx(pi) q[5];
        rz(pi/4) q[5];
        rz(pi/8) q[6];
        rz(pi/16) q[7];
        rz(pi/32) q[8];
        rz(pi/64) q[9];
        cx q[9],q[4];
        rz(-pi/64) q[4];
        cx q[9],q[4];
        rz(pi/64) q[4];
        cx q[8],q[4];
        rz(-pi/32) q[4];
        cx q[8],q[4];
        rz(pi/32) q[4];
        cx q[7],q[4];
        rz(-pi/16) q[4];
        cx q[7],q[4];
        rz(pi/16) q[4];
        cx q[6],q[4];
        rz(-pi/8) q[4];
        cx q[6],q[4];
        rz(pi/8) q[4];
        cx q[5],q[4];
        rz(-pi/4) q[4];
        cx q[5],q[4];
        rz(pi/4) q[4];
        ry(pi/2) q[4];
        rx(pi) q[4];
        rz(pi/4) q[4];
        rz(pi/8) q[5];
        rz(pi/16) q[6];
        rz(pi/32) q[7];
        rz(pi/64) q[8];
        rz(pi/128) q[9];
        cx q[9],q[3];
        rz(-pi/128) q[3];
        cx q[9],q[3];
        rz(pi/128) q[3];
        cx q[8],q[3];
        rz(-pi/64) q[3];
        cx q[8],q[3];
        rz(pi/64) q[3];
        cx q[7],q[3];
        rz(-pi/32) q[3];
        cx q[7],q[3];
        rz(pi/32) q[3];
        cx q[6],q[3];
        rz(-pi/16) q[3];
        cx q[6],q[3];
        rz(pi/16) q[3];
        cx q[5],q[3];
        rz(-pi/8) q[3];
        cx q[5],q[3];
        rz(pi/8) q[3];
        cx q[4],q[3];
        rz(-pi/4) q[3];
        cx q[4],q[3];
        rz(pi/4) q[3];
        ry(pi/2) q[3];
        rx(pi) q[3];
        rz(pi/4) q[3];
        rz(pi/8) q[4];
        rz(pi/16) q[5];
        rz(pi/32) q[6];
        rz(pi/64) q[7];
        rz(pi/128) q[8];
        rz(pi/256) q[9];
        cx q[9],q[2];
        rz(-pi/256) q[2];
        cx q[9],q[2];
        rz(pi/256) q[2];
        cx q[8],q[2];
        rz(-pi/128) q[2];
        cx q[8],q[2];
        rz(pi/128) q[2];
        cx q[7],q[2];
        rz(-pi/64) q[2];
        cx q[7],q[2];
        rz(pi/64) q[2];
        cx q[6],q[2];
        rz(-pi/32) q[2];
        cx q[6],q[2];
        rz(pi/32) q[2];
        cx q[5],q[2];
        rz(-pi/16) q[2];
        cx q[5],q[2];
        rz(pi/16) q[2];
        cx q[4],q[2];
        rz(-pi/8) q[2];
        cx q[4],q[2];
        rz(pi/8) q[2];
        cx q[3],q[2];
        rz(-pi/4) q[2];
        cx q[3],q[2];
        rz(pi/4) q[2];
        ry(pi/2) q[2];
        rx(pi) q[2];
        rz(pi/4) q[2];
        rz(pi/8) q[3];
        rz(pi/16) q[4];
        rz(pi/32) q[5];
        rz(pi/64) q[6];
        rz(pi/128) q[7];
        rz(pi/256) q[8];
        rz(pi/512) q[9];
        cx q[9],q[1];
        rz(-pi/512) q[1];
        cx q[9],q[1];
        rz(pi/512) q[1];
        cx q[8],q[1];
        rz(-pi/256) q[1];
        cx q[8],q[1];
        rz(pi/256) q[1];
        cx q[7],q[1];
        rz(-pi/128) q[1];
        cx q[7],q[1];
        rz(pi/128) q[1];
        cx q[6],q[1];
        rz(-pi/64) q[1];
        cx q[6],q[1];
        rz(pi/64) q[1];
        cx q[5],q[1];
        rz(-pi/32) q[1];
        cx q[5],q[1];
        rz(pi/32) q[1];
        cx q[4],q[1];
        rz(-pi/16) q[1];
        cx q[4],q[1];
        rz(pi/16) q[1];
        cx q[3],q[1];
        rz(-pi/8) q[1];
        cx q[3],q[1];
        rz(pi/8) q[1];
        cx q[2],q[1];
        rz(-pi/4) q[1];
        cx q[2],q[1];
        rz(pi/4) q[1];
        ry(pi/2) q[1];
        rx(pi) q[1];
        rz(pi/4) q[1];
        rz(pi/8) q[2];
        rz(pi/16) q[3];
        rz(pi/32) q[4];
        rz(pi/64) q[5];
        rz(pi/128) q[6];
        rz(pi/256) q[7];
        rz(pi/512) q[8];
        rz(pi/1024) q[9];
        cx q[9],q[0];
        rz(-pi/1024) q[0];
        cx q[9],q[0];
        rz(pi/1024) q[0];
        cx q[8],q[0];
        rz(-pi/512) q[0];
        cx q[8],q[0];
        rz(pi/512) q[0];
        cx q[7],q[0];
        rz(-pi/256) q[0];
        cx q[7],q[0];
        rz(pi/256) q[0];
        cx q[6],q[0];
        rz(-pi/128) q[0];
        cx q[6],q[0];
        rz(pi/128) q[0];
        cx q[5],q[0];
        rz(-pi/64) q[0];
        cx q[5],q[0];
        rz(pi/64) q[0];
        cx q[4],q[0];
        rz(-pi/32) q[0];
        cx q[4],q[0];
        rz(pi/32) q[0];
        cx q[3],q[0];
        rz(-pi/16) q[0];
        cx q[3],q[0];
        rz(pi/16) q[0];
        cx q[2],q[0];
        rz(-pi/8) q[0];
        cx q[2],q[0];
        rz(pi/8) q[0];
        cx q[1],q[0];
        rz(-pi/4) q[0];
        cx q[1],q[0];
        rz(pi/4) q[0];
        ry(pi/2) q[0];
        rx(pi) q[0];
        cx q[0],q[9];
        cx q[1],q[8];
        cx q[2],q[7];
        cx q[3],q[6];
        cx q[4],q[5];
        cx q[5],q[4];
        cx q[4],q[5];
        cx q[6],q[3];
        cx q[3],q[6];
        cx q[7],q[2];
        cx q[2],q[7];
        cx q[8],q[1];
        cx q[1],q[8];
        cx q[9],q[0];
        cx q[0],q[9];
    """
    circuit = QiskitCircuit.from_qasm_str(qasm)

    compiled_circuit = compile(
        circuit, target_gateset=["u3", "cx"], custom_passes=[MPSPass()]
    )

    fidelity = np.vdot(
        Statevector(circuit).data, Statevector(compiled_circuit).data
    )

    assert compiled_circuit.count_ops().get("cx", 0) == 0
    assert np.abs(np.round(fidelity, decimals=10)) == 1.0


def test_compile_with_target_gateset():
    """Test that the final circuit respects the user-defined gateset, no target device"""
    circuit = QiskitCircuit(2)
    circuit.cx(0, 1)
    circuit.h(0)

    target_gateset = {
        "ry",
        "rx",
        "cz",
    }
    result_circuit = compile(
        circuit,
        target_gateset=target_gateset,
    )

    assert set(op.name for op in result_circuit).issubset(target_gateset)


@pytest.mark.parametrize(
    "circuit_function", [qcnn_circuit, random_clifford_circuit]
)
@pytest.mark.parametrize("num_qubits", [6, 7, 8, 9, 10])
@pytest.mark.parametrize("seed", [1, 326, 5678, 12345])
def test_compilation_retains_gateset(circuit_function, num_qubits, seed):
    circuit = circuit_function(num_qubits, seed)
    transpiler = UCCDefault1()
    target_basis = transpiler.target_gateset
    transpiled_circuit = transpiler.run(circuit)
    dag = circuit_to_dag(transpiled_circuit)
    analysis_pass = GatesInBasis(basis_gates=target_basis)
    analysis_pass.run(dag)
    assert analysis_pass.property_set["all_gates_in_basis"]


# Test compilation accepts QASM circuits containing IF-ELSE
def test_compile_if_else():
    qasm = """
    OPENQASM 3;
    include "stdgates.inc";
    bit[3] data;
    bit[2] syndrome;
    qubit[3] q0;
    qubit[2] q1;

    syndrome[0] = measure q0[0];
    syndrome[1] = measure q1[1];
    if (syndrome[0]) {
        x q1[0];
    }
    if (syndrome[1]) {
        x q1[0];
    }
    """
    transpiled = compile(qasm, return_format="qiskit")
    assert isinstance(transpiled, QiskitCircuit)


@pytest.mark.parametrize(
    "circuit_function", [qcnn_circuit, random_clifford_circuit]
)
@pytest.mark.parametrize("num_qubits", [6, 7, 8, 9, 10, 15])
@pytest.mark.parametrize("seed", [1, 326, 5678, 12345])
def test_compiled_circuits_equivalent(circuit_function, num_qubits, seed):
    circuit = circuit_function(num_qubits, seed)
    transpiled = compile(circuit, return_format="qiskit")
    sv1 = Statevector(circuit)
    sv2 = Statevector(transpiled)
    assert sv1.equiv(sv2)



# Construct a custom compiler
import os

try:
    from qiskit.utils.parallel import default_num_processes
except ImportError:
    # Qiskit 1.0.0 doesn't have this function, so we make it ourselves
    from qiskit.utils.parallel import CPU_COUNT

    def default_num_processes():
        return CPU_COUNT


from qiskit.providers import Backend
from qiskit.transpiler import PassManager
from qiskit import user_config
from qiskit.transpiler import generate_preset_pass_manager
from qiskit.transpiler.passes import (
    ConsolidateBlocks,
    CollectCliffords,
    HighLevelSynthesis,
    HLSConfig,
    CommutativeCancellation,
    Collect2qBlocks,
    UnitarySynthesis,
    Optimize1qGatesDecomposition,
)
from typing import Optional


CONFIG = user_config.get_config()


class UCCDefault1:
    DEFAULT_GATESET = {"cx", "rz", "rx", "ry", "h"}

    def __init__(
        self,
        local_iterations: int = 1,
        target_backend: Optional[Backend] = None,
        target_gateset: Optional[set] = None,
    ):
        """
        Create a new instance of UCCDefault1 compiler

            Args:
                local_iterations (int): Number of times to run the local passes
                target_backend (qiskit.providers.Backend): (Optional) The       target backend device to compile the circuit for
                target_gateset (set[str]): (Optional) The gateset to compile the circuit to. e.g. {"cx", "rx",...}. `target_backend` takes precedence if it provides a basis gateset.

        If neither target_backend or target_gateset resolve to a gateset, defaults to {"cx", "rz", "rx", "ry", "h"}.

        """
        self.pass_manager = PassManager()
        self.target_backend = target_backend

        if self.target_backend is None:
            # If no backend is provided, use the provided gateset or default gateset
            self.target_gateset = (
                target_gateset
                if target_gateset is not None
                else self.DEFAULT_GATESET
            )
        elif hasattr(self.target_backend, "target") and hasattr(
            self.target_backend.target, "operation_names"
        ):
            # If a backend is provided, use its target's operation names as the gateset
            self.target_gateset = self.target_backend.target.operation_names
        else:
            raise ValueError(
                "Provided backend does not provide a target with operation names"
            )

        if self.target_backend is None:
            self._add_local_passes(local_iterations)
        else:
            self.pass_manager = generate_preset_pass_manager(
                optimization_level=3, backend=self.target_backend
            )

    @property
    def default_passes(self):
        return

    def _add_local_passes(self, local_iterations):
        for _ in range(local_iterations):
            self.pass_manager.append(Optimize1qGatesDecomposition())
            self.pass_manager.append(CommutativeCancellation())
            self.pass_manager.append(Collect2qBlocks())
            self.pass_manager.append(ConsolidateBlocks(force_consolidate=True))
            self.pass_manager.append(
                UnitarySynthesis(basis_gates=self.target_gateset)
            )
            # self.pass_manager.append(Optimize1qGatesDecomposition(basis=self._1q_basis))
            self.pass_manager.append(CollectCliffords())
            self.pass_manager.append(
                HighLevelSynthesis(hls_config=HLSConfig(clifford=["greedy"]))
            )

            # Add following passes if merging single qubit rotations that are interrupted by a commuting 2 qubit gate is desired
            # self.pass_manager.append(Optimize1qGatesSimpleCommutation(basis=self._1q_basis))

    def run(self, circuits, callback=None):
        """
        Run the pass manager on the given circuit(s).

            Args:
                circuits (QuantumCircuit or list[QuantumCircuit]): Circuit(s) to transpile
                callback: A callback function that will be called after each pass execution. The
                function will be called with 5 keyword arguments::

                    pass_ (Pass): the pass being run
                    dag (DAGCircuit): the dag output of the pass
                    time (float): the time to execute the pass
                    property_set (PropertySet): the property set
                    count (int): the index for the pass execution
        """
        return self.pass_manager.run(circuits, callback=callback)


def _get_trial_count(default_trials=5):
    if CONFIG.get("sabre_all_threads", None) or os.getenv(
        "QISKIT_SABRE_ALL_THREADS"
    ):
        return default_num_processes()
    return default_trials



from qbraid.programs.alias_manager import get_program_type_alias
from qbraid.transpiler import ConversionGraph
from qbraid.transpiler import transpile as translate
from qiskit import transpile as qiskit_transpile
from .transpilers.ucc_defaults import UCCDefault1

import sys
import warnings

# Specify the supported Python version range
REQUIRED_MAJOR = 3
MINOR_VERSION_MIN = 12
MINOR_VERSION_MAX = 13

current_major = sys.version_info.major
current_minor = sys.version_info.minor

if current_major != REQUIRED_MAJOR or not (
    MINOR_VERSION_MIN <= current_minor <= MINOR_VERSION_MAX
):
    warnings.warn(
        f"Warning: This package is designed for Python {REQUIRED_MAJOR}.{MINOR_VERSION_MIN}-{REQUIRED_MAJOR}.{MINOR_VERSION_MAX}. "
        f"You are using Python {current_major}.{current_minor}."
    )
supported_circuit_formats = ConversionGraph().nodes()


def compile(
    circuit,
    return_format="original",
    target_gateset=None,
    target_backend=None,
    custom_passes=None,
    callback=None,
):
    """Compiles the provided quantum `circuit` by translating it to a Qiskit
    circuit, transpiling it, and returning the optimized circuit in the
    specified `return_format`.

    Args:
        circuit (object): The quantum circuit to be compiled.
        return_format (str): The format in which your circuit will be returned.
            e.g., "TKET", "OpenQASM2". Check ``ucc.supported_circuit_formats``.
            Defaults to the format of the input circuit.
        target_gateset (set[str]): (optional) The gateset to compile the circuit to.
            e.g. {"cx", "rx",...}. Defaults to the gate set of the target device if available. If no `target_gateset` or ` target_backend` is provided, defaults to {"cx", "rz", "rx", "ry", "h"}.
        target_backend (qiskit.providers.backend): (optional)
            The target device  to compile the circuit for. Can be specified as a Qiskit backend. If None, all-to-all connectivity is assumed. If a `target_backend` is specified, `target_backend.operation_names` supercedes the `target_gateset`.
        custom_passes (list[qiskit.transpiler.TransformationPass]): (optional)
            A list of custom passes to apply after the default set
            of passes. Defaults to None.
        callback: A callback function that will be called after each pass execution. The
                function will be called with 5 keyword arguments::

                    pass_ (Pass): the pass being run
                    dag (DAGCircuit): the dag output of the pass
                    time (float): the time to execute the pass
                    property_set (PropertySet): the property set
                    count (int): the index for the pass execution

    Returns:
        object: The compiled circuit in the specified format.
    """
    if return_format == "original":
        return_format = get_program_type_alias(circuit)

    # Translate to Qiskit Circuit object
    qiskit_circuit = translate(circuit, "qiskit")

    # Initialize the UCCDefault1 compiler with the target device and gateset
    ucc_default1 = UCCDefault1(
        target_backend=target_backend, target_gateset=target_gateset
    )

    # Translate into the target device gateset first; no optimization
    basis_translated_circuit = qiskit_transpile(
        qiskit_circuit,
        basis_gates=ucc_default1.target_gateset,
        optimization_level=0,
    )

    if custom_passes is not None:
        ucc_default1.pass_manager.append(custom_passes)

    # Compile the circuit using the UCCDefault1 pass manager
    compiled_circuit = ucc_default1.run(
        basis_translated_circuit, callback=callback
    )

    # Translate the compiled circuit to the desired format
    final_result = translate(compiled_circuit, return_format)
    return final_result



import numpy as np
from numpy.typing import NDArray
import quimb.tensor as qtn  # type: ignore
from qiskit import QuantumCircuit  # type: ignore
from qiskit.quantum_info import Statevector  # type: ignore
from .mps_utils import calculate_entanglement_entropy_slope
import warnings
import logging

logger = logging.getLogger(__name__)


def gram_schmidt(matrix: NDArray[np.complex128]) -> NDArray[np.complex128]:
    """Perform Gram-Schmidt orthogonalization on the columns of a matrix
    to define the unitary block to encode the MPS.

    Notes
    -----
    If a column is (approximately) zero, it is replaced with a random vector.

    Args:
        matrix (NDArray[np.complex128]): Input matrix with complex entries.

    Returns:
        unitary (NDArray[np.complex128]): A unitary matrix with orthonormal columns
            derived from the input matrix. If a column is (approximately) zero, it
            is replaced with a random vector.
    """
    num_rows, num_columns = matrix.shape
    unitary = np.zeros((num_rows, num_columns), dtype=np.complex128)
    orthonormal_basis: list[NDArray[np.complex128]] = []

    for j in range(num_columns):
        column = matrix[:, j]

        # If column is (approximately) zero, replace with random
        if np.allclose(column, 0):
            column = np.random.uniform(-1, 1, num_rows)  # type: ignore
            if np.iscomplexobj(matrix):
                column = column + 1j * np.random.uniform(-1, 1, num_rows)

        # Gram-Schmidt orthogonalization
        for basis_vector in orthonormal_basis:
            column -= (basis_vector.conj().T @ column) * basis_vector

        # Handle near-zero vectors (linear dependence)
        norm = np.linalg.norm(column)
        if norm < 1e-12:
            is_complex = np.iscomplexobj(matrix)
            column = np.random.uniform(-1, 1, num_rows)  # type: ignore
            if is_complex:
                column += 1j * np.random.uniform(-1, 1, num_rows)
            for basis_vector in orthonormal_basis:
                column -= (basis_vector.conj().T @ column) * basis_vector

        unitary[:, j] = column / np.linalg.norm(column)
        orthonormal_basis.append(unitary[:, j])

    return unitary


class Sequential:
    def __init__(self, max_fidelity_threshold: float = 0.95) -> None:
        """Initialize the Sequential class.

        Args:
            max_fidelity_threshold (float): The maximum fidelity required, after
            which we can stop the encoding to save depth. Defaults to 0.95.
        """
        self.max_fidelity_threshold = max_fidelity_threshold

    def generate_layer(
        self, mps: qtn.MatrixProductState
    ) -> list[tuple[list[int], NDArray[np.complex128]]]:
        """Convert a Matrix Product State (MPS) to a circuit representation
        using a single unitary layer.

        Args:
            mps (qtn.MatrixProductState): The MPS to convert.

        Returns:
            unitary_layer (list[tuple[list[int], NDArray[np.complex128]]]): A list of
            tuples representing the unitary layer of the circuit.
                Each tuple contains:
                - A list of qubit indices (in LSB order) that the unitary acts on.
                - A unitary matrix (as a 2D NumPy array) that encodes the MPS.
        """
        num_sites = mps.L

        unitary_layer: list[tuple[list[int], NDArray[np.complex128]]] = []

        for i, tensor in enumerate(reversed(mps.arrays)):
            i = num_sites - i - 1

            # MPS representation uses 1D entanglement, thus we need to define
            # the range of the indices via the tensor shape
            # i.e., if q0 and q3 are entangled, then regardless of q1 and q2 being
            # entangled the entanglement range would be q0-q3
            if i == 0:
                d_right, d = tensor.shape
                tensor = tensor.reshape((1, d_right, d))
            if i == num_sites - 1:
                d_left, d = tensor.shape
                tensor = tensor.reshape((d_left, 1, d))

            tensor = np.swapaxes(tensor, 1, 2)

            # Combine the physical index and right-virtual index of the tensor to construct an isometry
            # matrix
            d_left, d, d_right = tensor.shape
            isometry = tensor.reshape((d * d_left, d_right))

            # Put qubit ordering in LSB (we provide this so that users can modify between LSB and MSB)
            # To put into MSB, comment the second line below
            qubits = reversed(range(i - int(np.ceil(np.log2(d_left))), i + 1))
            qubits = [abs(qubit - num_sites + 1) for qubit in qubits]  # type: ignore

            # Create all-zero matrix and add the isometry columns
            matrix = np.zeros(
                (isometry.shape[0], isometry.shape[0]), dtype=isometry.dtype
            )

            # Keep columns for which all ancillas are in the zero state
            matrix[:, : isometry.shape[1]] = isometry

            # Perform Gram-Schmidt orthogonalization to ensure the columns are orthonormal
            unitary = gram_schmidt(matrix)

            # Store the unitary layer for the circuit
            unitary_layer.append((qubits, unitary))  # type: ignore

        return unitary_layer

    def mps_to_circuit_approx(
        self,
        statevector: NDArray[np.complex128],
        max_num_layers: int,
        chi_max: int,
    ) -> QuantumCircuit:
        r"""Approximately encodes the MPS into a circuit via multiple layers
        of exact encoding of bond 2 truncated MPS.

        Whilst we can encode the MPS exactly in a single layer, we require
        $log(chi) + 1$ qubits for each tensor, which results in larger circuits.
        This function uses bond 2 which allows us to encode the MPS using one and
        two qubit gates, which results in smaller circuits, and easier to run on
        hardware.

        This is the core idea of Ran's paper [1].

        [1] https://arxiv.org/abs/1908.07958

        Args:
            statevector (NDArray[np.complex128]): The statevector to convert.
            max_num_layers (int): The number of layers to use in the circuit.

        Returns:
            QuantumCircuit: The generated quantum circuit that encodes the MPS.
        """
        mps = qtn.MatrixProductState.from_dense(statevector)
        mps: qtn.MatrixProductState = (
            qtn.tensor_1d_compress.tensor_network_1d_compress(
                mps, max_bond=chi_max
            )
        )  # type: ignore
        mps.permute_arrays()

        mps.compress(form="left", max_bond=chi_max)
        mps.left_canonicalize(normalize=True)

        compressed_mps = mps.copy(deep=True)
        disentangled_mps = mps.copy(deep=True)

        circuit = QuantumCircuit(mps.L, mps.L)

        unitary_layers = []

        # Initialize the zero state |00...0> to serve as comparison
        # for the disentangled MPS
        zero_state = np.zeros((2**mps.L,), dtype=np.complex128)
        zero_state[0] = 1.0  # |00...0> state

        # Ran's approach uses a iterative disentanglement of the MPS
        # where each layer compresses the MPS to a maximum bond dimension of 2
        # and applies the inverse of the layer to disentangle the MPS
        # After a few layers we are adequately close to |00...0> state
        # after which we can simply reverse the layers (no inverse) and apply them
        # to the |00...0> state to obtain the MPS state
        for layer_index in range(max_num_layers):
            # Compress the MPS from the previous layer to a maximum bond dimension of 2,
            # |ψ_k> -> |ψ'_k>
            compressed_mps = disentangled_mps.copy(deep=True)

            # Normalize improves fidelity of the encoding
            compressed_mps.normalize()
            compressed_mps.compress(form="left", max_bond=2)

            # Generate unitary layer
            unitary_layer = self.generate_layer(compressed_mps)
            unitary_layers.append(unitary_layer)

            # To update the MPS definition, apply the inverse of U_k to disentangle |ψ_k>,
            # |ψ_(k+1)> = inv(U_k) @ |ψ_k>
            for i, _ in enumerate(unitary_layer):
                inverse = unitary_layer[-(i + 1)][1].conj().T

                if inverse.shape[0] == 4:
                    disentangled_mps.gate_split_(inverse, (i - 1, i))
                else:
                    disentangled_mps.gate_(inverse, (i), contract=True)

            # Compress the disentangled MPS to a maximum bond dimension of chi_max
            # This is to ensure that the disentangled MPS does not grow too large
            # and improves the fidelity of the encoding
            disentangled_mps: qtn.MatrixProductState = (
                qtn.tensor_1d_compress.tensor_network_1d_compress(  # type: ignore
                    disentangled_mps, max_bond=chi_max
                )
            )

            fidelity = np.vdot(disentangled_mps.to_dense(), zero_state)

            if fidelity >= self.max_fidelity_threshold:
                # If the disentangled MPS is close enough to the zero state,
                # we can stop the disentanglement process
                logger.info(
                    f"Reached target fidelity {fidelity}. "
                    f"{layer_index + 1} layers used."
                )
                break

        if layer_index == max_num_layers - 1:
            logger.info(
                f"Reached fidelity {fidelity} with "
                f"maximum number of layers {max_num_layers}."
            )

        # The layers disentangle the MPS to a state close to |00...0>
        # inv(U_k) ... inv(U_1) |ψ> = |00...0>
        # So, we have to reverse the layers and apply them to the |00...0> state
        # to obtain the MPS state
        # |ψ> = U_1 ... U_k |00...0>
        unitary_layers.reverse()

        # Apply the unitary layers to the |00...0> state
        for unitary_layer in unitary_layers:
            for qubits, unitary in unitary_layer:
                circuit.unitary(unitary, qubits)

        return circuit

    @staticmethod
    def optimal_params(statevector: NDArray[np.complex128]) -> int:
        """Calculate the optimal number of layers and sweeps for the
        MPS encoding.

        Users should overwrite this static method to customize the definition
        of the number of layers and number of sweeps for their use-case.

        Args:
            statevector (NDArray[np.complex128]): The statevector to analyze.

        Returns:
            num_layers (int): The optimal maximum number of layers.
        """
        num_qubits = int(np.ceil(np.log2(len(statevector))))
        slope = calculate_entanglement_entropy_slope(statevector)

        # Entanglement entropy slope is between 0 and 1
        # Use a smooth transition between area-law (0 to 0.4) and volume-law (1)
        # The higher the slope, the more layers are needed
        num_layers = int((2 + 2 * slope) * num_qubits)

        return num_layers

    def __call__(
        self,
        statevector: NDArray[np.complex128],
    ) -> QuantumCircuit:
        """Call the instance to create the circuit that encodes the statevector.

        Args:
            statevector (NDArray[np.complex128]): The statevector to convert.

        Returns:
            QuantumCircuit: The generated quantum circuit.
        """
        num_qubits = int(np.ceil(np.log2(len(statevector))))

        # Single qubit statevector is optimal, and cannot be
        # further improved given depth of 1
        if num_qubits == 1:
            circuit = QuantumCircuit(1)
            circuit.initialize(statevector, [0])
            return circuit

        slope = calculate_entanglement_entropy_slope(statevector)
        if np.isclose(slope, 1, atol=0.1):
            warnings.warn(
                "Warning: The state is volume-law entangled. Compression may be too lossy."
            )

        max_num_layers = self.optimal_params(statevector)

        circuit = self.mps_to_circuit_approx(
            statevector, max_num_layers, 2**num_qubits
        )

        fidelity = np.vdot(Statevector(circuit).data, statevector)
        logger.info(
            f"Fidelity: {fidelity:.4f}, "
            f"Number of qubits: {num_qubits}, "
            f"Number of layers: {max_num_layers}, "
        )

        return circuit



from qiskit.converters import dag_to_circuit, circuit_to_dag
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.dagcircuit import DAGCircuit
from . import approx_compile


class MPSPass(TransformationPass):
    """Approximately compiles a quantum circuit using MPS IR to a staircase circuit
    using O(N) circuit depth.

    By default, the compiler uses `mps_sequential.py`, which provides a vanilla implementation
    of the encoding. To opt for the more performant version, install `qmprs` and the compiler
    will automatically use it.

    Below is the link to the repository:
    https://github.com/Qualition/qmprs
    """

    def __init__(self):
        super().__init__()

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        qiskit_circuit = dag_to_circuit(dag)

        return circuit_to_dag(approx_compile(qiskit_circuit))



import os
import sys

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../../../"))

from ucc._version import __version__
from datetime import datetime

project = "ucc"
current_year = datetime.now().year
copyright = f"2024–{current_year}, Unitary Foundation"
author = "Unitary Foundation"
directory_of_this_file = os.path.dirname(os.path.abspath(__file__))
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "myst_parser",
    "sphinx.ext.doctest",
    "sphinxcontrib.bibtex",
]

bibtex_bibfiles = ["refs.bib"]

myst_enable_extensions = [
    "linkify",
]
# Suppress warnings related to heading levels
suppress_warnings = ["myst.header"]

# Optionally, enable cross-referencing for `.rst` files as well
myst_crossref = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

source_suffix = [
    ".rst",
    ".md",
]  # Allow Sphinx to process both .rst and .md files


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = []



import importlib
import warnings
from qiskit import transpile as qiskit_transpile
from qiskit import QuantumCircuit
from qiskit.quantum_info import Statevector
import numpy as np

__all__ = [
    "calculate_entanglement_entropy_slope",
    "has_enough_memory",
    "qmprs_available",
    "MPS_Encoder",
    "approx_compile",
]

from .mps_utils import calculate_entanglement_entropy_slope, has_enough_memory

# Check if `qmprs` is installed for AQC compilation
qmprs_available = importlib.util.find_spec("qmprs") is not None
if qmprs_available:
    from .qmprs_compiler import QmprsCompiler as MPS_Encoder
else:
    from .mps_sequential import Sequential as MPS_Encoder


def approx_compile(circuit: QuantumCircuit) -> QuantumCircuit:
    """Compiles the qiskit circuit provided to approximately compile,
    and if the circuit state requires more memory than is available or
    the fidelity of aqc is lower than required it will return the original
    circuit.

    Args:
        qiskit_circuit (QuantumCircuit): The circuit to approximately compile.

    Returns:
        QuantumCircuit: The compiled or unchanged circuit.
    """
    if not qmprs_available:
        warnings.warn(
            "Warning: AQC compilation is requested, but `qmprs` is not installed. "
            "Falling back to vanilla sequential encoding."
        )

    if circuit.num_qubits == 1:
        return circuit

    # If the circuit's statevector IR requires more RAM than the user has,
    # ignore the compilation and return the inputted circuit as is
    has_memory, required_memory, available_memory = has_enough_memory(
        circuit.num_qubits
    )
    if not has_memory:
        warnings.warn(
            "Required memory to store statevector IR is more than available memory. \n"
            f"Required_memory: {required_memory} GB \n"
            f"Available memory to allocate to storing statevector IR: {available_memory} GB \n\n"
        )
        return circuit

    target_sv = Statevector(circuit).data
    aqc_circuit = MPS_Encoder()(target_sv)

    # Fallback protocol for low fidelity, which discards the compiled
    # circuit and returns the original one
    # TODO: This should be modified depending on maintainer notes
    fidelity = np.vdot(target_sv, Statevector(aqc_circuit).data)
    if fidelity < 0.8:
        warnings.warn(
            f"Warning: Fidelity {fidelity:.4f} is too low. Discarding compression."
        )
        return circuit

    # If the compiled circuit is deeper and has more cx than permitted, discard
    # the compilation
    aqc_transpiled = qiskit_transpile(
        aqc_circuit, basis_gates=["u3", "cx"], optimization_level=3
    )
    original_transpiled = qiskit_transpile(
        circuit, basis_gates=["u3", "cx"], optimization_level=3
    )

    aqc_cx_count = aqc_transpiled.count_ops().get("cx", 0)
    aqc_depth = aqc_transpiled.depth()

    original_cx_count = original_transpiled.count_ops().get("cx", 0)
    original_depth = original_transpiled.depth()

    # Fallback protocol for worse depth and cx counts, which discards
    # the compiled circuit and returns the original one
    # TODO: This should be modified depending on maintainer notes
    if (aqc_cx_count >= original_cx_count) and (aqc_depth >= original_depth):
        return circuit

    return aqc_circuit



from .compile import (
    compile as compile,
    supported_circuit_formats as supported_circuit_formats,
)

from .transpilers.ucc_defaults import UCCDefault1 as UCCDefault1
from ucc._version import __version__ as __version__



try:
    from bqskit.compiler import Compiler
    from bqskit.ext.qiskit.translate import bqskit_to_qiskit, qiskit_to_bqskit
    from bqskit.passes import (
        ForEachBlockPass,
        QuickPartitioner,
        LEAPSynthesisPass,
        UnfoldPass,
    )
except ImportError:
    raise ImportError(
        "bqskit is required to use BQSKitTransformationPass but is not a dependency for ucc, so it is not installed by default. You can install it from pypi e.g. 'pip install bqskit' or 'uv add bqskit'."
    )

import qiskit.transpiler.basepasses
import qiskit.converters


class BQSKitTransformationPass(
    qiskit.transpiler.basepasses.TransformationPass
):
    def __init__(self, bqskit_passes=None):
        super().__init__()
        if bqskit_passes is None:
            self.bqskit_passes = self.default_passes()
        else:
            self.bqskit_passes = bqskit_passes

    def default_passes(self):
        return [
            QuickPartitioner(3),
            ForEachBlockPass(
                LEAPSynthesisPass(), replace_filter="less-than-multi"
            ),
            UnfoldPass(),
        ]

    def run(self, dag):
        circuit = qiskit_to_bqskit(qiskit.converters.dag_to_circuit(dag))
        with Compiler() as compiler:
            circuit = compiler.compile(circuit, self.bqskit_passes)
        return qiskit.converters.circuit_to_dag(bqskit_to_qiskit(circuit))



import numpy as np
from numpy.typing import NDArray
import psutil
from qiskit.quantum_info import partial_trace, entropy


def calculate_entanglement_entropy_slope(
    state: NDArray[np.complex128],
) -> float:
    """Calculate the slope of the entanglement entropy as the
    subsystem size increases, which can indicate whether the
    entanglement is volume-law or area-law. This returns a float
    as opposed to a string to provide a more dynamic response.

    Args:
        state (NDArray[np.complex128]): The quantum state represented as a statevector.

    Returns:
        float: The entanglement entropy of the state.
    """
    n = int(np.ceil(np.log2(len(state))))

    entropies = []
    for k in range(1, n // 2 + 1):
        # Trace out rest of the qubits
        rho_A = partial_trace(state, list(range(k, n)))
        S = entropy(rho_A, base=2)
        entropies.append(S)

    # Check if the entropies form a straight line or a curve
    entropies = np.array(entropies[len(entropies) // 2 :])
    x = np.arange(1, len(entropies) + 1)

    # Linear regression: fit y = ax + b
    x_mean = np.mean(x)
    y_mean = np.mean(entropies)

    numerator = np.sum((x - x_mean) * (entropies - y_mean))
    denominator = np.sum((x - x_mean) ** 2)

    slope = numerator / denominator if denominator != 0 else 0

    return slope


def has_enough_memory(num_qubits: int) -> tuple[bool, float, float]:
    """Check if the available user RAM is enough to represent
    the statevector IR.

    Args:
        num_qubits (int): The number of qubits for the statevector.

    Returns:
        has_memory (bool): Whether the user has enough RAM.
        memory_required_gb (float): Amount of memory required to
            store the statevector IR in GB.
        available_memory_gb (float): Amount of free memory that
            can be dedicated to storing the statevector IR in GB.
    """
    available_memory_gb = psutil.virtual_memory().available
    available_memory_gb = available_memory_gb / 2**30

    # Calculate approximately how much memory the statevector
    # requires at worst-case (volume-law)
    # statevectors use np.complex128 which needs 16 bytes
    # Use half of the memory available to store the IR
    memory_required_gb = 2 ** (4 + num_qubits - 31)
    has_memory = memory_required_gb <= available_memory_gb

    return has_memory, memory_required_gb, available_memory_gb



import numpy as np
from numpy.typing import NDArray
from qmprs.synthesis.mps_encoding import Sequential as QmprsSequential  # type: ignore
from quick.circuit import QiskitCircuit
from qiskit import QuantumCircuit
from .mps_utils import calculate_entanglement_entropy_slope
import warnings
import logging

logger = logging.getLogger(__name__)


class QmprsCompiler:
    """Wrapper for `qmprs.synthesis.mps_encoding.Sequential` to approximately
    compile a statevector to a circuit using MPS encoding.

    For more information, see:
    https://github.com/Qualition/qmprs
    """

    def __init__(self, max_fidelity_threshold=0.9) -> None:
        """Initialize the QmprsCompiler with a target fidelity.

        Args:
            max_fidelity_threshold (float): The maximum fidelity required, after
            which we can stop the encoding to save depth. Defaults to 0.9.
        """
        self.sequential = QmprsSequential(QiskitCircuit)
        self.sequential.fidelity_threshold = max_fidelity_threshold

    @staticmethod
    def optimal_params(statevector: NDArray[np.complex128]) -> tuple[int, int]:
        """Calculate the optimal number of layers and sweeps for the
        MPS encoding.

        Users should overwrite this static method to customize the definition
        of the number of layers and number of sweeps for their use-case.

        Args:
            statevector (NDArray[np.complex128]): The statevector to analyze.

        Returns:
            tuple[int, int]: A tuple containing the number of layers and sweeps.
        """
        num_qubits = int(np.ceil(np.log2(len(statevector))))
        slope = calculate_entanglement_entropy_slope(statevector)

        # Entanglement entropy slope is between 0 and 1
        # Use a smooth transition between area-law (0 to 0.4) and volume-law (1)
        # The higher the slope, the more layers and sweeps are needed
        num_layers = int((2 + 1 * slope) * num_qubits)
        num_sweeps = int((10 + 20 * slope) * num_qubits)

        return num_layers, num_sweeps

    def __call__(self, statevector: NDArray[np.complex128]) -> QuantumCircuit:
        """Call the instance to create the circuit that encodes the statevector.

        Args:
            statevector (NDArray[np.complex128]): The statevector to convert.

        Returns:
            QuantumCircuit: The generated quantum circuit.
        """
        slope = calculate_entanglement_entropy_slope(statevector)
        if np.isclose(slope, 1, atol=0.1):
            warnings.warn(
                "Warning: The state is volume-law entangled. Compression may be too lossy."
            )

        num_qubits = int(np.ceil(np.log2(len(statevector))))

        # Single qubit statevector is optimal, and cannot be
        # further improved given depth of 1
        if num_qubits == 1:
            circuit = QuantumCircuit(1)
            circuit.initialize(statevector, [0])
            return circuit

        num_layers, num_sweeps = self.optimal_params(statevector)

        circuit = self.sequential.prepare_state(
            statevector=statevector,
            bond_dimension=2**num_qubits,
            num_layers=num_layers,
            num_sweeps=num_sweeps,
        )

        fidelity = np.vdot(circuit.get_statevector(), statevector)
        logger.info(
            f"Fidelity: {fidelity:.4f}, "
            f"Number of qubits: {num_qubits}, "
            f"Number of layers: {num_layers}, "
            f"Number of sweeps: {num_sweeps}"
        )

        return circuit.circuit
